# -*- coding: utf-8 -*-
import os

# PyQt5
from PyQt5 import QtCore, QtGui, QtWidgets
# NumPy
import numpy as np
# SciPy
from scipy.signal import find_peaks
# matplotlib
import matplotlib.pyplot as plt
import matplotlib.transforms as transforms
# AstroPy
import astropy.units as u
from astropy.table import Table
# specutils
from specutils import Spectrum1D
# drpy
from drpy import __version__ as version_drpy
from drpy.onedspec.center import refinePeaks
from drpy.modeling import Spline1D

from .__init__ import __version__ as version
from .ui import Ui_MainWindow, CheckBoxFileDialog, table_font
from .utils import _plotSpectrum, loadSpectrum, saveSpectrum

# Set plot parameters
plt.rcParams['axes.linewidth'] = 1
plt.rcParams['mathtext.fontset'] = 'stix'
plt.rcParams['font.family'] = 'STIXGeneral'


class MainWindow(QtWidgets.QMainWindow, Ui_MainWindow):


    def __init__(self):

        super().__init__()

        # Set up the user interface from Designer.
        self.setupUi(self)

        # Validate lineedit
        self.validateLineEdit()

        # Initialize lineedit
        self.initializeLineEdit()

        # Install event filter
        self.TableWidget_line.installEventFilter(self)

        # Connect up action `open`
        self.open_action.triggered.connect(self.load)
        # Connect up action `save`
        self.save_action.triggered.connect(self.save)
        # Connect up action `exit`
        self.exit_action.triggered.connect(self.close)
        # Connect up action `about`
        self.about_action.triggered.connect(self.about)
        # Connect up button `Find`
        self.Button_find.clicked.connect(self.find)
        # Connect up the switch button
        self.RectSwitch.clicked.connect(self.switchPlot)
        # Connect up button `Fit`
        self.Button_fit.clicked.connect(self.fitting)

        # Disable `save_action`
        self.save_action.setEnabled(False)
        # Disable group box `peak`
        self.setEnabled(group_box='peak', enable=False)
        # Disable group box `line`
        self.setEnabled(group_box='line', enable=False)
        # Disable group box `disp`
        self.setEnabled(group_box='disp', enable=False)
        
        # Tab order
        QtCore.QMetaObject.connectSlotsByName(self)
        self.setTabOrder(self.LineEdit_height, self.LineEdit_threshold)
        self.setTabOrder(self.LineEdit_threshold, self.LineEdit_distance)
        self.setTabOrder(self.LineEdit_distance, self.LineEdit_prominence)
        self.setTabOrder(self.LineEdit_prominence, self.LineEdit_width)
        self.setTabOrder(self.LineEdit_width, self.LineEdit_wlen)
        self.setTabOrder(self.LineEdit_wlen, self.LineEdit_rel)
        self.setTabOrder(self.LineEdit_rel, self.LineEdit_plateau)
        self.setTabOrder(self.LineEdit_plateau, self.Button_find)
        self.setTabOrder(self.Button_find, self.LineEdit_npieces)
        self.setTabOrder(self.LineEdit_npieces, self.RectSwitch)
        self.setTabOrder(self.RectSwitch, self.Button_fit)
        self.setTabOrder(self.Button_fit, self.TableWidget_line)
        
        # Flag
        self.isNewOpen = True


    def disableButtonFind(self, LineEdit):
        """Enable or disable the Button Find"""

        if len(LineEdit.text()) == 0:
            self.Button_find.setEnabled(False)

        else:
            self.Button_find.setEnabled(True)


    def disableButtonFit(self, LineEdit):
        """Enable or disable the Button Fit"""

        if len(LineEdit.text()) == 0:
            self.Button_fit.setEnabled(False)

        else:
            self.Button_fit.setEnabled(True)


    def validateLineEdit(self):

        # Validate peak parameters
        rx_0_1 = QtCore.QRegExp('^(1|(0(\\.\\d{1,3})?))$')
        rx_0_100 = QtCore.QRegExp('^(100|((\\d{1}|[1-9]{1}[0-9]{1})(\\.\\d{1,2})?))$')
        rx_1_100 = QtCore.QRegExp('^(100|(([1-9]{1}[0-9]?)(\\.\\d{1,2})?))$')
        self.LineEdit_height.setValidator(QtGui.QRegExpValidator(rx_0_1))
        self.LineEdit_threshold.setValidator(QtGui.QRegExpValidator(rx_0_1))
        self.LineEdit_distance.setValidator(QtGui.QRegExpValidator(rx_1_100))
        self.LineEdit_prominence.setValidator(QtGui.QRegExpValidator(rx_0_1))
        self.LineEdit_width.setValidator(QtGui.QRegExpValidator(rx_0_100))
        self.LineEdit_wlen.setValidator(QtGui.QRegExpValidator(rx_0_100))
        self.LineEdit_rel.setValidator(QtGui.QRegExpValidator(rx_0_1))
        self.LineEdit_plateau.setValidator(QtGui.QRegExpValidator(rx_0_100))
        # Not empty
        self.LineEdit_height.textChanged.connect(lambda: self.disableButtonFind(self.LineEdit_height))
        self.LineEdit_threshold.textChanged.connect(lambda: self.disableButtonFind(self.LineEdit_threshold))
        self.LineEdit_distance.textChanged.connect(lambda: self.disableButtonFind(self.LineEdit_distance))
        self.LineEdit_prominence.textChanged.connect(lambda: self.disableButtonFind(self.LineEdit_prominence))
        self.LineEdit_width.textChanged.connect(lambda: self.disableButtonFind(self.LineEdit_width))
        self.LineEdit_wlen.textChanged.connect(lambda: self.disableButtonFind(self.LineEdit_wlen))
        self.LineEdit_rel.textChanged.connect(lambda: self.disableButtonFind(self.LineEdit_rel))
        self.LineEdit_plateau.textChanged.connect(lambda: self.disableButtonFind(self.LineEdit_plateau))

        # Validate fit parameters
        rx_1_20 = QtCore.QRegExp('^(20|([2-9]{1})|(1\\d?))$')
        # rx_0_5 = QtCore.QRegExp('^(5|([0-4](\\.\\d{1,2})?))$')
        # rx_0_10 = QtCore.QRegExp('^(10|\\d{1})$')
        self.LineEdit_npieces.setValidator(QtGui.QRegExpValidator(rx_1_20))
        # self.LineEdit_sig_lower.setValidator(QtGui.QRegExpValidator(rx_0_5))
        # self.LineEdit_sig_upper.setValidator(QtGui.QRegExpValidator(rx_0_5))
        # self.LineEdit_maxiter.setValidator(QtGui.QRegExpValidator(rx_0_10))
        # Not empty
        self.LineEdit_npieces.textChanged.connect(lambda: self.disableButtonFit(self.LineEdit_npieces))
        # self.LineEdit_sig_lower.textChanged.connect(lambda: self.disableButtonFit(self.LineEdit_sig_lower))
        # self.LineEdit_sig_upper.textChanged.connect(lambda: self.disableButtonFit(self.LineEdit_sig_upper))
        # self.LineEdit_maxiter.textChanged.connect(lambda: self.disableButtonFit(self.LineEdit_maxiter))


    def initializeLineEdit(self):

        # Default peak parameters
        self.LineEdit_height.setText('0')
        self.LineEdit_threshold.setText('0')
        self.LineEdit_distance.setText('10')
        self.LineEdit_prominence.setText('0.05')
        self.LineEdit_width.setText('8')
        self.LineEdit_wlen.setText('20')
        self.LineEdit_rel.setText('0.5')
        self.LineEdit_plateau.setText('0')

        # Default unit
        self.LineEdit_unit.setText('Angstrom')

        # Default fit parameters
        self.LineEdit_npieces.setText('1')
        # self.LineEdit_sig_lower.setText('3')
        # self.LineEdit_sig_upper.setText('3')
        # self.LineEdit_maxiter.setText('1')


    def setEnabled(self, group_box, enable):
        """Enable or disable group box."""

        if group_box == 'peak':
            # Disable peak parameters
            self.LineEdit_height.setEnabled(enable)
            self.LineEdit_threshold.setEnabled(enable)
            self.LineEdit_distance.setEnabled(enable)
            self.LineEdit_prominence.setEnabled(enable)
            self.LineEdit_width.setEnabled(enable)
            self.LineEdit_wlen.setEnabled(enable)
            self.LineEdit_rel.setEnabled(enable)
            self.LineEdit_plateau.setEnabled(enable)
            # Disable button `Find`
            self.Button_find.setEnabled(enable)
        
        elif group_box == 'line':
            # Disable line table
            self.TableWidget_line.setEnabled(enable)
            # Disable unit
            self.LineEdit_unit.setEnabled(enable)

        elif group_box == 'disp':
            # Disable fit parameters
            self.LineEdit_npieces.setEnabled(enable)
            # self.LineEdit_sig_lower.setEnabled(enable)
            # self.LineEdit_sig_upper.setEnabled(enable)
            # self.LineEdit_maxiter.setEnabled(enable)
            # Disable the switch button
            self.RectSwitch.setEnabled(enable)
            # Disable button `Fit`
            self.Button_fit.setEnabled(enable)


    def setupCell(self, row, column, text='', isChecked=False):

        if column == 'wavelength':
            # Column `WAVELENGTH`
            item = QtWidgets.QTableWidgetItem(text)
            item.setTextAlignment(QtCore.Qt.AlignVCenter | QtCore.Qt.AlignRight)
            item.setFont(table_font)
            self.TableWidget_line.setItem(row, 0, item)

        elif column == 'mask':
            # Column `MASK`
            cell_widget = QtWidgets.QWidget()
            cell_layout = QtWidgets.QHBoxLayout(cell_widget)
            check_box = QtWidgets.QCheckBox()
            cell_widget.checkState = check_box.checkState
            if isChecked:
                check_box.setCheckState(QtCore.Qt.Checked)
            else:
                check_box.setCheckState(QtCore.Qt.Unchecked)
            cell_layout.addWidget(check_box)
            cell_layout.setAlignment(QtCore.Qt.AlignCenter)
            cell_layout.setContentsMargins(0, 0, 0, 0)
            self.TableWidget_line.setCellWidget(row, 1, cell_widget)


    def eventFilter(self, source, event):
        """copy, paste, cut, and delete"""

        if event.type() == QtCore.QEvent.KeyPress:

            if event.matches(QtGui.QKeySequence.Copy):
                self.copySelection()
                return True

            elif event.matches(QtGui.QKeySequence.Paste):
                self.pasteSelection()
                return True

            elif (event.matches(QtGui.QKeySequence.Delete) | (event.key() == QtCore.Qt.Key_Backspace)):
                self.deleteSelection()
                return True

            elif event.matches(QtGui.QKeySequence.Cut):
                self.copySelection()
                self.deleteSelection()
                return True

        return super(MainWindow, self).eventFilter(source, event)


    def copySelection(self):

        index_list = self.TableWidget_line.selectedIndexes()

        if index_list:

            isNone_list = [index.data() is None for index in index_list]
            index_list = sorted(index for index, isNone in zip(index_list, isNone_list) if not isNone)

            if index_list:

                copy_text = ''

                for index in index_list:
                    copy_text += index.data() + '\r\n'

                QtWidgets.qApp.clipboard().setText(copy_text)


    def pasteSelection(self):

        index_list = self.TableWidget_line.selectedIndexes()

        if index_list:

            isNone_list = [index.data() is None for index in index_list]
            index_list = sorted(index for index, isNone in zip(index_list, isNone_list) if not isNone)

            if index_list:

                if len(index_list) == (index_list[-1].row() - index_list[0].row() + 1):

                    paste_text = QtWidgets.qApp.clipboard().text()
                    paste_text_list = [text.strip('\r') for text in paste_text.split('\n')]

                    while paste_text_list:

                        if paste_text_list[-1] == '':
                            paste_text_list = paste_text_list[:-1]

                        else:
                            break

                    row0 = index_list[0].row()

                    for i in range(len(paste_text_list)):

                        if row0 + i < self.TableWidget_line.rowCount():
                            self.setupCell(row0 + i, column='wavelength', text=paste_text_list[i])


    def deleteSelection(self):

        index_list = self.TableWidget_line.selectedIndexes()

        if index_list:

            isNone_list = [index.data() is None for index in index_list]
            index_list = sorted(index for index, isNone in zip(index_list, isNone_list) if not isNone)

            if index_list:

                for index in index_list:
                    self.setupCell(index.row(), column='wavelength', text='')


    def plotSpectrum(self):
        """Plot spectrum.

        Input: ``self.index``, ``self.count``, ``self.unit_count``
        Output: None
        """

        if not self.isNewOpen:
            self.Figure_spectrum.clear()

        xlabel = 'spectral axis [pixel]'
        ylabel = 'flux'
        if self.unit_count.to_string():
            ylabel += f' [{self.unit_count.to_string()}]'

        ax = self.Figure_spectrum.add_subplot(111)
        _plotSpectrum(
            ax=ax, x=self.index, y=self.count, xlim=(self.index.min(), self.index.max()), 
            ylim=(-0.05, 1.15), xlabel=xlabel, ylabel=ylabel)
        self.Figure_spectrum.tight_layout(pad=0.2)
        self.FigureCanvas_spectrum.draw()


    def load(self, external=False):
        """Load spectrum.

        Input: ``self.file_name``, ``self.file_format`` and ``self.reverse`` (from here 
        or from command line)
        Output: ``self.index``, ``self.count`` ``self.unit_count``, ``self.header`` and ``self.peak_lib``

        Parameters
        ----------
        external : bool
            If `True`, a file name is provided in command line.
        """

        isAccepted = False

        if not external:

            file_dialog = CheckBoxFileDialog(check_box_text='reverse')

            file_dialog.setNameFilters(
                ['FITS (*.fits *.fit *.FITS *.FIT)', 'Enhanced CSV Files (*.ecsv)', 'ASCII Files (*.*)'])

            if self.file_name is not None:
                file_dialog.selectFile(self.file_name)

            if self.file_format == 'ascii':
                file_dialog.selectNameFilter('ASCII Files (*.*)')

            elif self.file_format == 'ecsv':
                file_dialog.selectNameFilter('Enhanced CSV Files (*.ecsv)')

            elif self.file_format == 'fits':
                file_dialog.selectNameFilter('FITS (*.fits *.fit *.FITS *.FIT)')

            if self.reverse is not None:
                file_dialog.check_box.setChecked(self.reverse)

            if file_dialog.exec_() == QtWidgets.QDialog.Accepted:
                
                isAccepted = True

                # ``file_name``
                self.file_name = file_dialog.selectedUrls()[0].toLocalFile()

                # ``file_format``
                extension = os.path.splitext(self.file_name)[1]
                if file_dialog.selectedNameFilter().startswith('Enhanced'):
                    self.file_format = 'ecsv'

                elif file_dialog.selectedNameFilter().startswith('FITS'):
                    self.file_format = 'fits'

                else:
                    self.file_format = 'ascii'

                # ``reverse``
                self.reverse = file_dialog.check_box.checkState() == QtCore.Qt.Checked

        if external | isAccepted:

            self.index, self.count, self.unit_count, self.header = loadSpectrum(
                file_name=self.file_name, file_format=self.file_format, 
                reverse=self.reverse)

            self.peak_lib = dict()

            # Plot
            self.plotSpectrum()

            if self.isNewOpen:

                # Enable group box `peak`
                self.setEnabled(group_box='peak', enable=True)
                # Set flag `isNewOpen`
                self.isNewOpen = False

            else:

                # Initialize the table
                self.TableWidget_line.setRowCount(0)
                # Clear group box `fit`
                self.clearFitting()


    def save(self):
        """Save calibrated spectrum

        Input: ``self.spectrum``, ``self.peak_table``
        Output: None
        """

        file_dialog = CheckBoxFileDialog(check_box_text='Enhanced CSV Files (*.ecsv)')
        file_dialog.setAcceptMode(QtWidgets.QFileDialog.AcceptSave)
        file_dialog.setNameFilters(['FITS (*.fits *.fit *.FITS *.FIT)'])
        file_dialog.selectFile(os.path.splitext(self.file_name)[0])

        if file_dialog.exec_() == QtWidgets.QDialog.Accepted:

            # ``file_name``
            file_name = file_dialog.selectedUrls()[0].toLocalFile()
            extension = os.path.splitext(file_name)[1]

            if extension not in ['.fits', '.fit', '.FITS', '.FIT']:
                file_name += '.fits'

            # also save as Enhanced CSV Files (*.ecsv) or not
            saveECSV = file_dialog.check_box.checkState() == QtCore.Qt.Checked

            saveSpectrum(
                file_name=file_name, spectrum=self.spectrum, peak_table=self.peak_table, 
                saveECSV=saveECSV)
            
            self.statusBar.showMessage(f'Saved to {file_name} successfully.', 2000)

    # todo: add urls
    def about(self):

        text = (
            f'Version: {version}\n' +\
            'Developer: Ruining ZHAO\n' +\
            f'Date: 2022-12-03\n' +\
            f'drpy: version {version_drpy}\n' +\
            'todo-list:\n' + \
            '1) Bugs to be fixed.\n' +\
            '2) Exceptions to be caught.'
        )
        about_message_box = QtWidgets.QMessageBox()
        about_message_box.setIcon(QtWidgets.QMessageBox.Information)
        about_message_box.setText('Wavelength Calibrator')
        about_message_box.setInformativeText(text)
        about_message_box.setWindowTitle('About')
        about_message_box.exec_()

    # todo: dynamic xlabel and ylabel
    def find(self):
        """Find peaks.

        Input: ``self.peak_lib``, ``self.count``
        Output: ``self.peaks``, ``self.properties``, ``self.peak_lib``
        """
    
        n_row_old = self.TableWidget_line.rowCount()
        # Get old contents
        if n_row_old > 0:

            for i in range(n_row_old):
                self.peak_lib[int(self.peaks[i])] = (
                    self.TableWidget_line.item(i, 0).text(),
                    self.TableWidget_line.cellWidget(i, 1).checkState()
                )

        # Get peak parameters
        height = float(self.LineEdit_height.text())
        threshold = float(self.LineEdit_threshold.text())
        distance = float(self.LineEdit_distance.text())
        prominence = float(self.LineEdit_prominence.text())
        width = float(self.LineEdit_width.text())
        wlen = float(self.LineEdit_wlen.text())
        rel = float(self.LineEdit_rel.text())
        plateau = float(self.LineEdit_plateau.text())

        # Find peaks
        peaks, properties = find_peaks(
            x=self.count, height=height, threshold=threshold, distance=distance, 
            prominence=prominence, width=width, wlen=wlen, rel_height=rel, 
            plateau_size=plateau)

        # Refine peaks
        if peaks.shape[0] > 0:
            self.peaks, self.properties, _ = refinePeaks(
                spectrum=self.count, peaks=peaks, properties=properties)

        else:
            self.peaks = peaks
            self.properties = properties

        self.TableWidget_line.setRowCount(len(self.peaks))
        peak_lib_keys = self.peak_lib.keys()

        for i, peak in enumerate(self.peaks):

            if int(peak) in peak_lib_keys:
                text, isChecked = self.peak_lib[int(peak)]
                # Column `WAVELENGTH`
                self.setupCell(row=i, column='wavelength', text=text)
                # Column `MASK`
                self.setupCell(row=i, column='mask', isChecked=isChecked)

            else:
                # Column `WAVELENGTH`
                self.setupCell(row=i, column='wavelength', text='')
                # Column `MASK`
                self.setupCell(row=i, column='mask', isChecked=False)

            # Row height
            self.TableWidget_line.verticalHeader().setSectionResizeMode(
                i, QtWidgets.QHeaderView.ResizeToContents)
        
        n_row_new = self.TableWidget_line.rowCount()

        if n_row_new > 0:
            self.setEnabled(group_box='line', enable=True)
            self.setEnabled(group_box='disp', enable=True)

        else:
            self.clearFitting()

        # Plot
        xlabel = 'spectral axis [pixel]'
        ylabel = 'flux'
        if self.unit_count.to_string():
            ylabel += f' [{self.unit_count.to_string()}]'

        self.Figure_spectrum.clear()
        ax = self.Figure_spectrum.add_subplot(111)
        trans = transforms.blended_transform_factory(
            ax.transData, ax.transAxes)
        for i, peak in enumerate(self.peaks):
            ax.plot([peak, peak], [0.0, 0.87], 'r--', transform=trans, lw=0.8)
            ax.plot([peak, peak], [0.94, 1.0], 'r--', transform=trans, lw=0.8)
            ax.annotate(
                f'{i + 1:d}', xy=(peak, 0.9), xycoords=trans, ha='center', va='center', 
                fontsize=12, color='r', annotation_clip=True)
        _plotSpectrum(
            ax=ax, x=self.index, y=self.count, xlim=(self.index.min(), self.index.max()), 
            ylim=(-0.05, 1.15), xlabel=xlabel, ylabel=ylabel)
        self.Figure_spectrum.tight_layout(pad=0.2)
        self.FigureCanvas_spectrum.draw()


    def clearFitting(self):
        """Clear fitting plot."""
        
        # Clear plot
        if hasattr(self, 'ax_fit'):
            del self.ax_fit

        self.Figure_fit.clear()
        self.FigureCanvas_fit.draw()

        # Disable group box `disp`
        self.setEnabled(group_box='line', enable=False)
        self.setEnabled(group_box='disp', enable=False)

        # Disable Button `Save`
        self.save_action.setEnabled(False)


    def fitting(self):
        """
        Input: ``self.index``, ``self.wave``, ``self.mask``, ``self.wave_peaks``
        
        Output: ``self.spectrum``, ``self.peak_table``, ``self.show_residual``

        Raise
        -----
        1) unit error
        2) fitting error
        """

        # Get fit parameters
        n_piece = int(self.LineEdit_npieces.text())

        if not hasattr(self, 'show_residual'):
            self.show_residual = self.RectSwitch.isChecked()
        
        # Get labelled lines
        n_row = self.TableWidget_line.rowCount()
        self.wave_input = np.zeros(n_row)
        self.mask_input = np.zeros(n_row, dtype=bool)

        for i in range(n_row):

            try:
                w = float(self.TableWidget_line.item(i, 0).text())
                m = bool(self.TableWidget_line.cellWidget(i, 1).checkState())

            except:
                w = np.nan
                m = True

            self.wave_input[i] = w
            self.mask_input[i] = m

        # Fit cubic spline function
        spl, self.residual, _, _, _ = Spline1D(
            x=self.peaks, y=self.wave_input, weight=None, mask=self.mask_input, 
            order=3, n_piece=n_piece, n_iter=0, sigma_lower=None, 
            sigma_upper=None, grow=False, use_relative=False)

        wavelength = spl(self.index)

        self.rms = round(
            np.sqrt((self.residual[~self.mask_input]**2).sum() / (~self.mask_input).sum()), 3)

        unit_wavelength = u.Unit(self.LineEdit_unit.text())

        self.spectrum = Spectrum1D(
            spectral_axis=(wavelength * unit_wavelength), 
            flux=(self.count * self.unit_count), 
            meta={'header': self.header})
        
        # Plot
        self.plotFitting()

        meta_peak = {
            'EXTNAME': ('peak', 'name of the extension'), 
            'PEAKUSED': ((~self.mask_input).sum(), 'number of peaks used'), 
            'PEAKDETE': (self.wave_input.shape[0], 'number of peaks detected'), 
            'PIECE': (n_piece, 'number of cubic spline pieces'), 
            'RMS': (self.rms, 'root mean squared of fitting'), 
        }

        self.peak_table = Table(
            data=[self.peaks[~self.mask_input] * u.pixel, 
                  self.properties['peak_heights'][~self.mask_input] * self.unit_count, 
                  self.properties['left_bases'][~self.mask_input] * u.pixel, 
                  self.properties['right_bases'][~self.mask_input] * u.pixel, 
                  self.wave_input[~self.mask_input] * unit_wavelength], 
            names=('peaks', 'heights', 'left_bases', 'right_bases', 'spectral_axis'), 
            meta=meta_peak,
        )

        # Enable Button `Save`
        self.save_action.setEnabled(True)


    def plotFitting(self):
        """self.wave_index -> self.spectrum.spectral_axis

        Input: ``self.show_residual``, ``self.wave_input``, ``self.peaks``, 
        ``self.mask_input``, ``self.spectrum``, ``self.residual``, ``self.rms``
        Output: None
        """

        xlabel = 'spectral axis'
        ylabel = 'residuals'
        if self.spectrum.spectral_axis.unit.to_string():
            xlabel += f' [{self.spectrum.spectral_axis.unit.to_string()}]'
            ylabel += f' [{self.spectrum.spectral_axis.unit.to_string()}]'

        self.Figure_fit.clear()
        self.ax_fit = self.Figure_fit.add_subplot(111)

        if not self.show_residual:

            self.ax_fit.plot(
                self.wave_input[self.mask_input], self.peaks[self.mask_input], 'x', 
                c='grey', ms=8)
            self.ax_fit.plot(
                self.wave_input[~self.mask_input], self.peaks[~self.mask_input], 'x', 
                c='red', ms=8)
            self.ax_fit.plot(
                self.spectrum.spectral_axis.value, self.index, 'k-', lw=0.8)
            self.ax_fit.set_ylabel('dispersion axis [pixel]', fontsize=12)

        else:

            self.ax_fit.plot(
                self.wave_input[~self.mask_input], self.residual[~self.mask_input], 
                'x', c='red', ms=8)
            self.ax_fit.axhline(y=0, ls='--', color='k', lw=0.8)

            for i, peak in enumerate(self.peaks):

                if self.mask_input[i]:
                    self.ax_fit.annotate(
                        f'{i + 1:d}', (self.wave_input[i], 0), xycoords='data', 
                        ha='center', va='center', fontsize=12, color='grey')

                else:
                    self.ax_fit.annotate(
                        f'{i + 1:d}', (self.wave_input[i], 0), xycoords='data', 
                        ha='center', va='center', fontsize=12, color='r')

            self.ax_fit.set_ylabel(ylabel, fontsize=12)

        self.ax_fit.tick_params(
            which='major', direction='in', top=True, right=True, length=5, width=1.5, 
            labelsize=12)
        self.ax_fit.tick_params(
            which='minor', direction='in', top=True, right=True, length=3, width=1.5, 
            labelsize=12)
        self.ax_fit.set_xlim(
            self.spectrum.spectral_axis.value.min(), self.spectrum.spectral_axis.value.max())
        self.ax_fit.set_xlabel(xlabel, fontsize=12)
        self.ax_fit.annotate(
            f'$N={(~self.mask_input).sum()}/{self.wave_input.shape[0]}' + '$, $\\rm{RMS}=' + f'{self.rms}$',
            xy=(0.98, 0.1), xycoords='axes fraction', ha='right', fontsize=12)
        self.Figure_fit.tight_layout(pad=0.2, h_pad=0)
        self.FigureCanvas_fit.draw()


    def switchPlot(self):

        self.show_residual = self.RectSwitch.isChecked()

        if hasattr(self, 'ax_fit'):
            self.plotFitting()


# -*- coding: utf-8 -*-
import os, sys, argparse, warnings

# PyQt5
from PyQt5 import QtWidgets

from .mainwindow import MainWindow
    
def main():
    """Command line tool."""
    
    # UI
    app = QtWidgets.QApplication([])
    app.setStyle('Windows')
    # Setup
    main_window = MainWindow()
    # Show
    main_window.show()

    parser = argparse.ArgumentParser()
    parser.add_argument(
        '-n', '--name', default=None, 
        help='Name of the input file'
    )
    parser.add_argument(
        '-f', '--format', default=None, 
        help='Format of the input file (`ascii`, `ecsv`, or `fits`).'
    )
    parser.add_argument(
        '-r', '--reverse', action='store_true', default=False, 
        help='Reverse or not.'
    )

    # Parse
    args = parser.parse_args()
    main_window.file_name = args.name
    main_window.file_format = args.format
    main_window.reverse = args.reverse

    # warnings.filterwarnings('error')

    # Load spectrum if ``name`` and ``format`` is provided
    if main_window.file_name is not None:

        main_window.file_name = os.path.abspath(main_window.file_name)

        if main_window.file_format is not None:

            main_window.load(external=True)

    sys.exit(app.exec_())

if __name__ == '__main__':
    main()


# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'wavecal.ui'
#
# Created by: PyQt5 UI code generator 5.9
#
# WARNING! All changes made in this file will be lost!

# PyQt5
from PyQt5 import QtCore, QtGui, QtWidgets
# matplotlib
from matplotlib.backends.qt_compat import QtWidgets
from matplotlib.backends.backend_qtagg import (
    FigureCanvas, NavigationToolbar2QT as NavigationToolbar)
import matplotlib.pyplot as plt

# switch
from .widgets import RectSwitch
# fonts
from .fonts import table_font


class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.MinimumExpanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        # sizePolicy.setHeightForWidth(MainWindow.sizePolicy().hasHeightForWidth())

        # The main window
        MainWindow.resize(1200, 800)
        MainWindow.setSizePolicy(sizePolicy)
        MainWindow.setMinimumSize(QtCore.QSize(1200, 800))
        MainWindow.setObjectName("MainWindow")

        # The menu bar (Added by RNZ 11/06/2021)
        self.menu_bar = MainWindow.menuBar()
        self.menu_bar.setNativeMenuBar(False)
        self.file_menu = self.menu_bar.addMenu("&File")
        self.open_action = QtWidgets.QAction("Open File...", MainWindow)
        self.open_action.setShortcut("Ctrl+O")
        self.save_action = QtWidgets.QAction("Save As...", MainWindow)
        self.save_action.setShortcut("Ctrl+S")
        self.exit_action = QtWidgets.QAction("Exit", MainWindow)
        self.exit_action.setShortcut("Ctrl+Q")
        self.file_menu.addAction(self.open_action)
        self.file_menu.addSeparator()
        self.file_menu.addAction(self.save_action)
        self.file_menu.addSeparator()
        self.file_menu.addAction(self.exit_action)
        self.help_menu = self.menu_bar.addMenu("&Help")
        self.about_action = QtWidgets.QAction("About", MainWindow)
        self.help_menu.addAction(self.about_action)

        # The central widget
        self.centralwidget = QtWidgets.QWidget(MainWindow)
        self.centralwidget.setObjectName("centralwidget")

        # The grid layout
        self.gridLayout = QtWidgets.QGridLayout(self.centralwidget)
        self.gridLayout.setObjectName("gridLayout")
        
        # Group box `peak`
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.MinimumExpanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        # sizePolicy.setHeightForWidth(self.GroupBox_peak.sizePolicy().hasHeightForWidth())
        self.GroupBox_peak = QtWidgets.QGroupBox(self.centralwidget)
        self.GroupBox_peak.setSizePolicy(sizePolicy)
        self.GroupBox_peak.setMinimumSize(QtCore.QSize(100, 276))
        self.GroupBox_peak.setObjectName("GroupBox_peak")

        # Horizontal layout 1
        self.HorizontalLayout_1 = QtWidgets.QHBoxLayout(self.GroupBox_peak)
        self.HorizontalLayout_1.setObjectName("HorizontalLayout_1")

        # Vertical layout `spectrum`
        self.VerticalLayout_spectrum = QtWidgets.QVBoxLayout()
        self.VerticalLayout_spectrum.setObjectName("VerticalLayout_spectrum")

        # Canvas `spectrum`
        self.Figure_spectrum = plt.figure()
        self.FigureCanvas_spectrum = FigureCanvas(self.Figure_spectrum)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.MinimumExpanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.FigureCanvas_spectrum.sizePolicy().hasHeightForWidth())
        self.FigureCanvas_spectrum.setSizePolicy(sizePolicy)
        self.FigureCanvas_spectrum.setMinimumSize(QtCore.QSize(100, 276))
        self.FigureCanvas_spectrum.setObjectName("FigureCanvas_spectrum")
        
        # Tool bar `spectrum`
        self.ToolBar_spectrum = NavigationToolbar(self.FigureCanvas_spectrum, MainWindow)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.ToolBar_spectrum.sizePolicy().hasHeightForWidth())
        self.ToolBar_spectrum.setSizePolicy(sizePolicy)
        self.ToolBar_spectrum.setMinimumSize(QtCore.QSize(100, 30))
        self.ToolBar_spectrum.setMaximumSize(QtCore.QSize(16777215, 30))
        self.ToolBar_spectrum.setObjectName("ToolBar_spectrum")
        
        # Add to vertical layout `spectrum`
        self.VerticalLayout_spectrum.addWidget(self.ToolBar_spectrum)
        self.VerticalLayout_spectrum.addWidget(self.FigureCanvas_spectrum)

        # Add to horizontal layout 1
        self.HorizontalLayout_1.addLayout(self.VerticalLayout_spectrum)

        # Vertical layout `peak`
        self.VerticalLayout_peak = QtWidgets.QVBoxLayout()
        self.VerticalLayout_peak.setContentsMargins(5, -1, 5, -1)
        self.VerticalLayout_peak.setSpacing(5)
        self.VerticalLayout_peak.setObjectName("VerticalLayout_peak")

        # Label `peak`
        self.Label_peak = QtWidgets.QLabel(self.GroupBox_peak)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.Label_peak.sizePolicy().hasHeightForWidth())
        self.Label_peak.setSizePolicy(sizePolicy)
        self.Label_peak.setMinimumSize(QtCore.QSize(180, 25))
        self.Label_peak.setMaximumSize(QtCore.QSize(180, 25))
        self.Label_peak.setObjectName("Label_peak")
        # Add to vertical layout `peak`
        self.VerticalLayout_peak.addWidget(self.Label_peak)

        # Line `peak`
        self.Line_peak = QtWidgets.QFrame(self.GroupBox_peak)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.Line_peak.sizePolicy().hasHeightForWidth())
        self.Line_peak.setSizePolicy(sizePolicy)
        self.Line_peak.setMinimumSize(QtCore.QSize(180, 0))
        self.Line_peak.setMaximumSize(QtCore.QSize(180, 16777215))
        self.Line_peak.setFrameShape(QtWidgets.QFrame.HLine)
        self.Line_peak.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.Line_peak.setObjectName("Line_peak")
        # Add to vertical layout `peak`
        self.VerticalLayout_peak.addWidget(self.Line_peak)

        # Form layout `peak`
        self.FormLayout_peak = QtWidgets.QFormLayout()
        self.FormLayout_peak.setSizeConstraint(QtWidgets.QLayout.SetFixedSize)
        self.FormLayout_peak.setFieldGrowthPolicy(QtWidgets.QFormLayout.AllNonFixedFieldsGrow)
        self.FormLayout_peak.setLabelAlignment(QtCore.Qt.AlignLeading|QtCore.Qt.AlignLeft|QtCore.Qt.AlignVCenter)
        self.FormLayout_peak.setFormAlignment(QtCore.Qt.AlignLeading|QtCore.Qt.AlignLeft|QtCore.Qt.AlignTop)
        self.FormLayout_peak.setContentsMargins(0, -1, 0, -1)
        self.FormLayout_peak.setHorizontalSpacing(10)
        self.FormLayout_peak.setObjectName("FormLayout_peak")
        # 1.1 height label
        self.Label_height = QtWidgets.QLabel(self.GroupBox_peak)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.Label_height.sizePolicy().hasHeightForWidth())
        self.Label_height.setSizePolicy(sizePolicy)
        self.Label_height.setMinimumSize(QtCore.QSize(97, 20))
        self.Label_height.setMaximumSize(QtCore.QSize(97, 20))
        self.Label_height.setObjectName("Label_height")
        self.FormLayout_peak.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.Label_height)
        # 1.1 height lineedit
        self.LineEdit_height = QtWidgets.QLineEdit(self.GroupBox_peak)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.LineEdit_height.sizePolicy().hasHeightForWidth())
        self.LineEdit_height.setSizePolicy(sizePolicy)
        self.LineEdit_height.setMinimumSize(QtCore.QSize(73, 20))
        self.LineEdit_height.setMaximumSize(QtCore.QSize(73, 20))
        self.LineEdit_height.setObjectName("LineEdit_height")
        self.FormLayout_peak.setWidget(0, QtWidgets.QFormLayout.FieldRole, self.LineEdit_height)
        # 1.2 threshold label
        self.Label_threshold = QtWidgets.QLabel(self.GroupBox_peak)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.Label_threshold.sizePolicy().hasHeightForWidth())
        self.Label_threshold.setSizePolicy(sizePolicy)
        self.Label_threshold.setMinimumSize(QtCore.QSize(97, 20))
        self.Label_threshold.setMaximumSize(QtCore.QSize(97, 20))
        self.Label_threshold.setObjectName("Label_threshold")
        self.FormLayout_peak.setWidget(1, QtWidgets.QFormLayout.LabelRole, self.Label_threshold)
        # 1.2 threshold lineedit
        self.LineEdit_threshold = QtWidgets.QLineEdit(self.GroupBox_peak)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.LineEdit_threshold.sizePolicy().hasHeightForWidth())
        self.LineEdit_threshold.setSizePolicy(sizePolicy)
        self.LineEdit_threshold.setMinimumSize(QtCore.QSize(73, 20))
        self.LineEdit_threshold.setMaximumSize(QtCore.QSize(73, 20))
        self.LineEdit_threshold.setObjectName("LineEdit_threshold")
        self.FormLayout_peak.setWidget(1, QtWidgets.QFormLayout.FieldRole, self.LineEdit_threshold)
        # 1.3 distance label
        self.Label_distance = QtWidgets.QLabel(self.GroupBox_peak)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.Label_distance.sizePolicy().hasHeightForWidth())
        self.Label_distance.setSizePolicy(sizePolicy)
        self.Label_distance.setMinimumSize(QtCore.QSize(97, 20))
        self.Label_distance.setMaximumSize(QtCore.QSize(97, 20))
        self.Label_distance.setObjectName("Label_distance")
        self.FormLayout_peak.setWidget(2, QtWidgets.QFormLayout.LabelRole, self.Label_distance)
        # 1.3 distance lineedit
        self.LineEdit_distance = QtWidgets.QLineEdit(self.GroupBox_peak)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.LineEdit_distance.sizePolicy().hasHeightForWidth())
        self.LineEdit_distance.setSizePolicy(sizePolicy)
        self.LineEdit_distance.setMinimumSize(QtCore.QSize(73, 20))
        self.LineEdit_distance.setMaximumSize(QtCore.QSize(73, 20))
        self.LineEdit_distance.setBaseSize(QtCore.QSize(0, 0))
        self.LineEdit_distance.setObjectName("LineEdit_distance")
        self.FormLayout_peak.setWidget(2, QtWidgets.QFormLayout.FieldRole, self.LineEdit_distance)
        # 1.4 prominence label
        self.Label_prominence = QtWidgets.QLabel(self.GroupBox_peak)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.Label_prominence.sizePolicy().hasHeightForWidth())
        self.Label_prominence.setSizePolicy(sizePolicy)
        self.Label_prominence.setMinimumSize(QtCore.QSize(97, 20))
        self.Label_prominence.setMaximumSize(QtCore.QSize(97, 20))
        self.Label_prominence.setObjectName("Label_prominence")
        self.FormLayout_peak.setWidget(3, QtWidgets.QFormLayout.LabelRole, self.Label_prominence)
        # 1.4 prominence lineedit
        self.LineEdit_prominence = QtWidgets.QLineEdit(self.GroupBox_peak)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.LineEdit_prominence.sizePolicy().hasHeightForWidth())
        self.LineEdit_prominence.setSizePolicy(sizePolicy)
        self.LineEdit_prominence.setMinimumSize(QtCore.QSize(73, 20))
        self.LineEdit_prominence.setMaximumSize(QtCore.QSize(73, 20))
        self.LineEdit_prominence.setObjectName("LineEdit_prominence")
        self.FormLayout_peak.setWidget(3, QtWidgets.QFormLayout.FieldRole, self.LineEdit_prominence)
        # 1.5 width label
        self.Label_width = QtWidgets.QLabel(self.GroupBox_peak)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.Label_width.sizePolicy().hasHeightForWidth())
        self.Label_width.setSizePolicy(sizePolicy)
        self.Label_width.setMinimumSize(QtCore.QSize(97, 20))
        self.Label_width.setMaximumSize(QtCore.QSize(97, 20))
        self.Label_width.setObjectName("Label_width")
        self.FormLayout_peak.setWidget(4, QtWidgets.QFormLayout.LabelRole, self.Label_width)
        # 1.5 width lineedit
        self.LineEdit_width = QtWidgets.QLineEdit(self.GroupBox_peak)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.LineEdit_width.sizePolicy().hasHeightForWidth())
        self.LineEdit_width.setSizePolicy(sizePolicy)
        self.LineEdit_width.setMinimumSize(QtCore.QSize(73, 20))
        self.LineEdit_width.setMaximumSize(QtCore.QSize(73, 20))
        self.LineEdit_width.setObjectName("LineEdit_width")
        self.FormLayout_peak.setWidget(4, QtWidgets.QFormLayout.FieldRole, self.LineEdit_width)
        # 1.6 wlen label
        self.Label_wlen = QtWidgets.QLabel(self.GroupBox_peak)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.Label_wlen.sizePolicy().hasHeightForWidth())
        self.Label_wlen.setSizePolicy(sizePolicy)
        self.Label_wlen.setMinimumSize(QtCore.QSize(97, 20))
        self.Label_wlen.setMaximumSize(QtCore.QSize(97, 20))
        self.Label_wlen.setObjectName("Label_wlen")
        self.FormLayout_peak.setWidget(5, QtWidgets.QFormLayout.LabelRole, self.Label_wlen)
        # 1.6 wlen lineedit
        self.LineEdit_wlen = QtWidgets.QLineEdit(self.GroupBox_peak)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.LineEdit_wlen.sizePolicy().hasHeightForWidth())
        self.LineEdit_wlen.setSizePolicy(sizePolicy)
        self.LineEdit_wlen.setMinimumSize(QtCore.QSize(73, 20))
        self.LineEdit_wlen.setMaximumSize(QtCore.QSize(73, 20))
        self.LineEdit_wlen.setObjectName("LineEdit_wlen")
        self.FormLayout_peak.setWidget(5, QtWidgets.QFormLayout.FieldRole, self.LineEdit_wlen)
        # 1.7 rel label
        self.Label_rel = QtWidgets.QLabel(self.GroupBox_peak)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.Label_rel.sizePolicy().hasHeightForWidth())
        self.Label_rel.setSizePolicy(sizePolicy)
        self.Label_rel.setMinimumSize(QtCore.QSize(97, 20))
        self.Label_rel.setMaximumSize(QtCore.QSize(97, 20))
        self.Label_rel.setObjectName("Label_rel")
        self.FormLayout_peak.setWidget(6, QtWidgets.QFormLayout.LabelRole, self.Label_rel)
        # 1.7 rel lineedit
        self.LineEdit_rel = QtWidgets.QLineEdit(self.GroupBox_peak)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.LineEdit_rel.sizePolicy().hasHeightForWidth())
        self.LineEdit_rel.setSizePolicy(sizePolicy)
        self.LineEdit_rel.setMinimumSize(QtCore.QSize(73, 20))
        self.LineEdit_rel.setMaximumSize(QtCore.QSize(73, 20))
        self.LineEdit_rel.setObjectName("LineEdit_rel")
        self.FormLayout_peak.setWidget(6, QtWidgets.QFormLayout.FieldRole, self.LineEdit_rel)
        # 1.8 plateau label
        self.Label_plateau = QtWidgets.QLabel(self.GroupBox_peak)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.Label_plateau.sizePolicy().hasHeightForWidth())
        self.Label_plateau.setSizePolicy(sizePolicy)
        self.Label_plateau.setMinimumSize(QtCore.QSize(97, 20))
        self.Label_plateau.setMaximumSize(QtCore.QSize(97, 20))
        self.Label_plateau.setObjectName("Label_plateau")
        self.FormLayout_peak.setWidget(7, QtWidgets.QFormLayout.LabelRole, self.Label_plateau)
        # 1.8 plateau lineedit
        self.LineEdit_plateau = QtWidgets.QLineEdit(self.GroupBox_peak)
        self.LineEdit_plateau.setEnabled(True)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.LineEdit_plateau.sizePolicy().hasHeightForWidth())
        self.LineEdit_plateau.setSizePolicy(sizePolicy)
        self.LineEdit_plateau.setMinimumSize(QtCore.QSize(73, 20))
        self.LineEdit_plateau.setMaximumSize(QtCore.QSize(73, 20))
        self.LineEdit_plateau.setObjectName("LineEdit_plateau")
        self.FormLayout_peak.setWidget(7, QtWidgets.QFormLayout.FieldRole, self.LineEdit_plateau)
        # 1.9 help label (Added by RNZ 11/06/2021)
        self.Label_help_find_peak = QtWidgets.QLabel(self.GroupBox_peak)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.Label_help_find_peak.sizePolicy().hasHeightForWidth())
        self.Label_help_find_peak.setSizePolicy(sizePolicy)
        self.Label_help_find_peak.setMinimumSize(QtCore.QSize(73, 20))
        self.Label_help_find_peak.setMaximumSize(QtCore.QSize(73, 20))
        self.Label_help_find_peak.setObjectName("Label_help_find_peak")
        self.FormLayout_peak.setWidget(8, QtWidgets.QFormLayout.LabelRole, self.Label_help_find_peak)
        
        # Add to vertical layout `peak`
        self.VerticalLayout_peak.addLayout(self.FormLayout_peak)
        
        # Button `Find`
        self.Button_find = QtWidgets.QPushButton(self.GroupBox_peak)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.Button_find.sizePolicy().hasHeightForWidth())
        self.Button_find.setSizePolicy(sizePolicy)
        self.Button_find.setMinimumSize(QtCore.QSize(180, 20))
        self.Button_find.setMaximumSize(QtCore.QSize(180, 20))
        self.Button_find.setObjectName("Button_find")
        # Add to vertical layout `peak`
        self.VerticalLayout_peak.addWidget(self.Button_find)
        
        # Add to horizontal layout 1
        self.HorizontalLayout_1.addLayout(self.VerticalLayout_peak)
        self.HorizontalLayout_1.setStretch(0, 1)
        
        # Add to the grid layout
        self.gridLayout.addWidget(self.GroupBox_peak, 0, 0, 1, 1)

        # Group box `line`
        self.GroupBox_line = QtWidgets.QGroupBox(self.centralwidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.MinimumExpanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.GroupBox_line.sizePolicy().hasHeightForWidth())
        self.GroupBox_line.setSizePolicy(sizePolicy)
        self.GroupBox_line.setMinimumSize(QtCore.QSize(230, 700))
        self.GroupBox_line.setMaximumSize(QtCore.QSize(230, 16777215))
        self.GroupBox_line.setObjectName("GroupBox_line")

        # Vertical layout 2
        self.VerticalLayout_2 = QtWidgets.QVBoxLayout(self.GroupBox_line)
        self.VerticalLayout_2.setContentsMargins(5, -1, 5, -1)
        self.VerticalLayout_2.setSpacing(10)
        self.VerticalLayout_2.setObjectName("VerticalLayout_2")

        # Table widget `line`
        self.TableWidget_line = QtWidgets.QTableWidget(self.GroupBox_line)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.MinimumExpanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.TableWidget_line.sizePolicy().hasHeightForWidth())
        self.TableWidget_line.setSizePolicy(sizePolicy)
        self.TableWidget_line.setMinimumSize(QtCore.QSize(210, 600))
        self.TableWidget_line.setMaximumSize(QtCore.QSize(210, 16777215))
        self.TableWidget_line.setObjectName("TableWidget_line")
        self.TableWidget_line.setColumnCount(2)
        self.TableWidget_line.setRowCount(0)

        # Header 0
        item = QtWidgets.QTableWidgetItem()
        item.setTextAlignment(QtCore.Qt.AlignCenter)
        item.setFont(table_font)
        self.TableWidget_line.setHorizontalHeaderItem(0, item)

        # Header 1
        item = QtWidgets.QTableWidgetItem()
        item.setTextAlignment(QtCore.Qt.AlignCenter)
        item.setFont(table_font)
        self.TableWidget_line.setHorizontalHeaderItem(1, item)
        
        # Resize mode
        self.TableWidget_line.horizontalHeader().setSectionResizeMode(0, QtWidgets.QHeaderView.Stretch)
        self.TableWidget_line.horizontalHeader().setSectionResizeMode(1, QtWidgets.QHeaderView.ResizeToContents)

        # Add to vertical layout 2
        self.VerticalLayout_2.addWidget(self.TableWidget_line)

        # Form layout `unit`
        self.FormLayout_unit = QtWidgets.QFormLayout()
        self.FormLayout_unit.setSizeConstraint(QtWidgets.QLayout.SetFixedSize)
        self.FormLayout_unit.setFieldGrowthPolicy(QtWidgets.QFormLayout.AllNonFixedFieldsGrow)
        self.FormLayout_unit.setLabelAlignment(QtCore.Qt.AlignLeading|QtCore.Qt.AlignLeft|QtCore.Qt.AlignVCenter)
        self.FormLayout_unit.setFormAlignment(QtCore.Qt.AlignLeading|QtCore.Qt.AlignLeft|QtCore.Qt.AlignTop)
        self.FormLayout_unit.setContentsMargins(2, -1, 0, -1)
        self.FormLayout_unit.setHorizontalSpacing(10)
        self.FormLayout_unit.setObjectName("FormLayout_unit")

        # unit label
        self.Label_unit = QtWidgets.QLabel(self.GroupBox_line)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.Label_unit.sizePolicy().hasHeightForWidth())
        self.Label_unit.setSizePolicy(sizePolicy)
        self.Label_unit.setMinimumSize(QtCore.QSize(108, 20))
        self.Label_unit.setMaximumSize(QtCore.QSize(108, 20))
        self.Label_unit.setObjectName("Label_unit")
        self.FormLayout_unit.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.Label_unit)
        # unit lineedit
        self.LineEdit_unit = QtWidgets.QLineEdit(self.GroupBox_line)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.LineEdit_unit.sizePolicy().hasHeightForWidth())
        self.LineEdit_unit.setSizePolicy(sizePolicy)
        self.LineEdit_unit.setMinimumSize(QtCore.QSize(90, 20))
        self.LineEdit_unit.setMaximumSize(QtCore.QSize(90, 20))
        self.LineEdit_unit.setObjectName("LineEdit_unit")
        self.FormLayout_unit.setWidget(0, QtWidgets.QFormLayout.FieldRole, self.LineEdit_unit)

        # Add to vertical layout 2
        self.VerticalLayout_2.addLayout(self.FormLayout_unit)
        self.VerticalLayout_2.setStretch(0, 1)

        # Add to the grid layout
        self.gridLayout.addWidget(self.GroupBox_line, 0, 1, 2, 1)

        # Group box `disp`
        self.GroupBox_disp = QtWidgets.QGroupBox(self.centralwidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.MinimumExpanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.GroupBox_disp.sizePolicy().hasHeightForWidth())
        self.GroupBox_disp.setSizePolicy(sizePolicy)
        self.GroupBox_disp.setMinimumSize(QtCore.QSize(100, 276))
        self.GroupBox_disp.setObjectName("GroupBox_disp")
        
        # Horizontal layout 3
        self.HorizontalLayout_3 = QtWidgets.QHBoxLayout(self.GroupBox_disp)
        self.HorizontalLayout_3.setObjectName("HorizontalLayout_3")

        # Vertical layout `fit`
        self.VerticalLayout_fit = QtWidgets.QVBoxLayout()
        self.VerticalLayout_fit.setObjectName("VerticalLayout_fit")

        # Canvas `fit`
        self.Figure_fit = plt.figure()
        self.FigureCanvas_fit = FigureCanvas(self.Figure_fit)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.MinimumExpanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.FigureCanvas_fit.sizePolicy().hasHeightForWidth())
        self.FigureCanvas_fit.setSizePolicy(sizePolicy)
        self.FigureCanvas_fit.setMinimumSize(QtCore.QSize(100, 276))
        self.FigureCanvas_fit.setObjectName("FigureCanvas_fit")

        # Tool bar `fit`
        self.ToolBar_fit = NavigationToolbar(self.FigureCanvas_fit, MainWindow)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.ToolBar_fit.sizePolicy().hasHeightForWidth())
        self.ToolBar_fit.setSizePolicy(sizePolicy)
        self.ToolBar_fit.setMinimumSize(QtCore.QSize(100, 30))
        self.ToolBar_fit.setMaximumSize(QtCore.QSize(16777215, 30))
        self.ToolBar_fit.setObjectName("ToolBar_fit")
        
        # Add to Vertical layout `fit`
        self.VerticalLayout_fit.addWidget(self.ToolBar_fit)
        self.VerticalLayout_fit.addWidget(self.FigureCanvas_fit)
        
        # Add to horizontal layout 3
        self.HorizontalLayout_3.addLayout(self.VerticalLayout_fit)

        # Vertical layout `disp`
        self.VerticalLayout_disp = QtWidgets.QVBoxLayout()
        self.VerticalLayout_disp.setSpacing(5)
        self.VerticalLayout_disp.setContentsMargins(5, -1, 5, -1)
        self.VerticalLayout_disp.setObjectName("VerticalLayout_disp")

        # Label `disp`
        self.Label_disp = QtWidgets.QLabel(self.GroupBox_disp)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.Label_disp.sizePolicy().hasHeightForWidth())
        self.Label_disp.setSizePolicy(sizePolicy)
        self.Label_disp.setMinimumSize(QtCore.QSize(180, 25))
        self.Label_disp.setMaximumSize(QtCore.QSize(180, 25))
        self.Label_disp.setObjectName("Label_disp")
        # Add to vertical layout `disp`
        self.VerticalLayout_disp.addWidget(self.Label_disp)
        
        # Line `disp`
        self.Line_disp = QtWidgets.QFrame(self.GroupBox_disp)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.Line_disp.sizePolicy().hasHeightForWidth())
        self.Line_disp.setSizePolicy(sizePolicy)
        self.Line_disp.setMinimumSize(QtCore.QSize(180, 0))
        self.Line_disp.setMaximumSize(QtCore.QSize(180, 16777215))
        self.Line_disp.setFrameShape(QtWidgets.QFrame.HLine)
        self.Line_disp.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.Line_disp.setObjectName("Line_disp")
        # Add to vertical layout `disp`
        self.VerticalLayout_disp.addWidget(self.Line_disp)

        # Form layout `disp`
        self.FormLayout_disp = QtWidgets.QFormLayout()
        self.FormLayout_disp.setSizeConstraint(QtWidgets.QLayout.SetFixedSize)
        self.FormLayout_disp.setFieldGrowthPolicy(QtWidgets.QFormLayout.AllNonFixedFieldsGrow)
        self.FormLayout_disp.setLabelAlignment(QtCore.Qt.AlignLeading|QtCore.Qt.AlignLeft|QtCore.Qt.AlignVCenter)
        self.FormLayout_disp.setFormAlignment(QtCore.Qt.AlignLeading|QtCore.Qt.AlignLeft|QtCore.Qt.AlignTop)
        self.FormLayout_disp.setContentsMargins(0, -1, 0, -1)
        self.FormLayout_disp.setHorizontalSpacing(17)
        self.FormLayout_disp.setObjectName("FormLayout_disp")
        # 1.1 npieces label
        self.Label_npieces = QtWidgets.QLabel(self.GroupBox_disp)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.Label_npieces.sizePolicy().hasHeightForWidth())
        self.Label_npieces.setSizePolicy(sizePolicy)
        self.Label_npieces.setMinimumSize(QtCore.QSize(90, 20))
        self.Label_npieces.setMaximumSize(QtCore.QSize(90, 20))
        self.Label_npieces.setObjectName("Label_npieces")
        self.FormLayout_disp.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.Label_npieces)
        # 1.1 npieces lineedit
        self.LineEdit_npieces = QtWidgets.QLineEdit(self.GroupBox_disp)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.LineEdit_npieces.sizePolicy().hasHeightForWidth())
        self.LineEdit_npieces.setSizePolicy(sizePolicy)
        self.LineEdit_npieces.setMinimumSize(QtCore.QSize(73, 20))
        self.LineEdit_npieces.setMaximumSize(QtCore.QSize(73, 20))
        self.LineEdit_npieces.setObjectName("LineEdit_npieces")
        self.FormLayout_disp.setWidget(0, QtWidgets.QFormLayout.FieldRole, self.LineEdit_npieces)
        # # 1.2 sig_lower label
        # self.Label_sig_lower = QtWidgets.QLabel(self.GroupBox_disp)
        # sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        # sizePolicy.setHorizontalStretch(0)
        # sizePolicy.setVerticalStretch(0)
        # sizePolicy.setHeightForWidth(self.Label_sig_lower.sizePolicy().hasHeightForWidth())
        # self.Label_sig_lower.setSizePolicy(sizePolicy)
        # self.Label_sig_lower.setMinimumSize(QtCore.QSize(90, 20))
        # self.Label_sig_lower.setMaximumSize(QtCore.QSize(90, 20))
        # self.Label_sig_lower.setObjectName("Label_sig_lower")
        # self.FormLayout_disp.setWidget(1, QtWidgets.QFormLayout.LabelRole, self.Label_sig_lower)
        # # 1.2 sig_lower lineedit
        # self.LineEdit_sig_lower = QtWidgets.QLineEdit(self.GroupBox_disp)
        # sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        # sizePolicy.setHorizontalStretch(0)
        # sizePolicy.setVerticalStretch(0)
        # sizePolicy.setHeightForWidth(self.LineEdit_sig_lower.sizePolicy().hasHeightForWidth())
        # self.LineEdit_sig_lower.setSizePolicy(sizePolicy)
        # self.LineEdit_sig_lower.setMinimumSize(QtCore.QSize(73, 20))
        # self.LineEdit_sig_lower.setMaximumSize(QtCore.QSize(73, 20))
        # self.LineEdit_sig_lower.setObjectName("LineEdit_sig_lower")
        # self.FormLayout_disp.setWidget(1, QtWidgets.QFormLayout.FieldRole, self.LineEdit_sig_lower)
        # # 1.3 sig_upper label
        # self.Label_sig_upper = QtWidgets.QLabel(self.GroupBox_disp)
        # sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        # sizePolicy.setHorizontalStretch(0)
        # sizePolicy.setVerticalStretch(0)
        # sizePolicy.setHeightForWidth(self.Label_sig_upper.sizePolicy().hasHeightForWidth())
        # self.Label_sig_upper.setSizePolicy(sizePolicy)
        # self.Label_sig_upper.setMinimumSize(QtCore.QSize(90, 20))
        # self.Label_sig_upper.setMaximumSize(QtCore.QSize(90, 20))
        # self.Label_sig_upper.setObjectName("Label_sig_upper")
        # self.FormLayout_disp.setWidget(2, QtWidgets.QFormLayout.LabelRole, self.Label_sig_upper)
        # # 1.3 sig_upper lineedit
        # self.LineEdit_sig_upper = QtWidgets.QLineEdit(self.GroupBox_disp)
        # sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        # sizePolicy.setHorizontalStretch(0)
        # sizePolicy.setVerticalStretch(0)
        # sizePolicy.setHeightForWidth(self.LineEdit_sig_upper.sizePolicy().hasHeightForWidth())
        # self.LineEdit_sig_upper.setSizePolicy(sizePolicy)
        # self.LineEdit_sig_upper.setMinimumSize(QtCore.QSize(73, 20))
        # self.LineEdit_sig_upper.setMaximumSize(QtCore.QSize(73, 20))
        # self.LineEdit_sig_upper.setObjectName("LineEdit_sig_upper")
        # self.FormLayout_disp.setWidget(2, QtWidgets.QFormLayout.FieldRole, self.LineEdit_sig_upper)
        # # 1.4 maxiter label
        # self.Label_maxiter = QtWidgets.QLabel(self.GroupBox_disp)
        # sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        # sizePolicy.setHorizontalStretch(0)
        # sizePolicy.setVerticalStretch(0)
        # sizePolicy.setHeightForWidth(self.Label_maxiter.sizePolicy().hasHeightForWidth())
        # self.Label_maxiter.setSizePolicy(sizePolicy)
        # self.Label_maxiter.setMinimumSize(QtCore.QSize(90, 20))
        # self.Label_maxiter.setMaximumSize(QtCore.QSize(90, 20))
        # self.Label_maxiter.setObjectName("Label_maxiter")
        # self.FormLayout_disp.setWidget(3, QtWidgets.QFormLayout.LabelRole, self.Label_maxiter)
        # # 1.4 maxiter lineedit
        # self.LineEdit_maxiter = QtWidgets.QLineEdit(self.GroupBox_disp)
        # sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        # sizePolicy.setHorizontalStretch(0)
        # sizePolicy.setVerticalStretch(0)
        # sizePolicy.setHeightForWidth(self.LineEdit_maxiter.sizePolicy().hasHeightForWidth())
        # self.LineEdit_maxiter.setSizePolicy(sizePolicy)
        # self.LineEdit_maxiter.setMinimumSize(QtCore.QSize(73, 20))
        # self.LineEdit_maxiter.setMaximumSize(QtCore.QSize(73, 20))
        # self.LineEdit_maxiter.setObjectName("LineEdit_maxiter")
        # self.FormLayout_disp.setWidget(3, QtWidgets.QFormLayout.FieldRole, self.LineEdit_maxiter)
        # 1.5 the switch label
        self.Label_switch = QtWidgets.QLabel(self.GroupBox_disp)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.Label_switch.sizePolicy().hasHeightForWidth())
        self.Label_switch.setSizePolicy(sizePolicy)
        self.Label_switch.setMinimumSize(QtCore.QSize(90, 20))
        self.Label_switch.setMaximumSize(QtCore.QSize(90, 20))
        self.Label_switch.setObjectName("Label_switch")
        self.FormLayout_disp.setWidget(1, QtWidgets.QFormLayout.LabelRole, self.Label_switch)
        # 1.5 the switch
        self.RectSwitch = RectSwitch(self.GroupBox_disp, margin=-9, thumb_width=0.4, duration=150)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.RectSwitch.sizePolicy().hasHeightForWidth())
        self.RectSwitch.setSizePolicy(sizePolicy)
        self.RectSwitch.setMinimumSize(QtCore.QSize(73, 20))
        self.RectSwitch.setMaximumSize(QtCore.QSize(73, 20))
        self.RectSwitch.setObjectName("RectSwitch")
        self.FormLayout_disp.setWidget(1, QtWidgets.QFormLayout.FieldRole, self.RectSwitch)

        # Add to vertical layout `disp`
        self.VerticalLayout_disp.addLayout(self.FormLayout_disp)
        
        # Button `fit`
        self.Button_fit = QtWidgets.QPushButton(self.GroupBox_disp)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.Button_fit.sizePolicy().hasHeightForWidth())
        self.Button_fit.setSizePolicy(sizePolicy)
        self.Button_fit.setMinimumSize(QtCore.QSize(180, 20))
        self.Button_fit.setMaximumSize(QtCore.QSize(180, 20))
        self.Button_fit.setObjectName("Button_fit")
        # Add to vertical layout `disp`
        self.VerticalLayout_disp.addWidget(self.Button_fit)
        
        # Add to horizontal layout 3
        self.HorizontalLayout_3.addLayout(self.VerticalLayout_disp)
        self.HorizontalLayout_3.setStretch(0, 1)
        
        # Add to the grid layout
        self.gridLayout.addWidget(self.GroupBox_disp, 1, 0, 1, 1)
        self.gridLayout.setRowMinimumHeight(0, 1)
        self.gridLayout.setRowMinimumHeight(1, 1)
        self.gridLayout.setColumnStretch(0, 1)

        # Add to the main window
        MainWindow.setCentralWidget(self.centralwidget)
        
        # The status bar
        self.statusBar = QtWidgets.QStatusBar(MainWindow)
        self.statusBar.setObjectName("statusBar")
        # Add to the main window
        MainWindow.setStatusBar(self.statusBar)

        # Retranslate
        self.retranslateUi(MainWindow)

    def retranslateUi(self, MainWindow):
        _translate = QtCore.QCoreApplication.translate
        MainWindow.setWindowTitle(_translate("MainWindow", "Wavelength Calibrator"))
        self.GroupBox_peak.setTitle(_translate("MainWindow", "Peak Detection"))
        self.Label_peak.setText(_translate("MainWindow", "<html><head/><body><p><span style=\" font-size:10pt;\">Peak parameters</span></p></body></html>"))
        self.Label_height.setText(_translate("MainWindow", "<html><head/><body><p><span style=\" font-size:10pt;\">height</span></p></body></html>"))
        self.Label_threshold.setText(_translate("MainWindow", "<html><head/><body><p><span style=\" font-size:10pt;\">threshold</span></p></body></html>"))
        self.Label_distance.setText(_translate("MainWindow", "<html><head/><body><p><span style=\" font-size:10pt;\">distance</span></p></body></html>"))
        self.Label_prominence.setText(_translate("MainWindow", "<html><head/><body><p><span style=\" font-size:10pt;\">prominence</span></p></body></html>"))
        self.Label_width.setText(_translate("MainWindow", "<html><head/><body><p><span style=\" font-size:10pt;\">width</span></p></body></html>"))
        self.Label_wlen.setText(_translate("MainWindow", "<html><head/><body><p><span style=\" font-size:10pt;\">wlen</span></p></body></html>"))
        self.Label_rel.setText(_translate("MainWindow", "<html><head/><body><p><span style=\" font-size:10pt;\">rel_height</span></p></body></html>"))
        self.Label_plateau.setText(_translate("MainWindow", "<html><head/><body><p><span style=\" font-size:10pt;\">plateau_size</span></p></body></html>"))
        self.Label_help_find_peak.setText(_translate("MainWindow", "<a href=\"https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.find_peaks.html\">help</a>"))
        self.Label_help_find_peak.setOpenExternalLinks(True)
        self.Button_find.setText(_translate("MainWindow", "Find"))
        
        self.GroupBox_line.setTitle(_translate("MainWindow", "Line Identification"))
        self.Label_unit.setText(_translate("MainWindow", "<html><head/><body><p><span style=\" font-size:10pt;\">wavelength unit</span></p></body></html>"))
        item = self.TableWidget_line.horizontalHeaderItem(0)
        item.setText(_translate("MainWindow", "WAVELENGTH"))
        item = self.TableWidget_line.horizontalHeaderItem(1)
        item.setText(_translate("MainWindow", "MASK"))
        
        self.GroupBox_disp.setTitle(_translate("MainWindow", "Dispersion Solution"))
        self.Label_disp.setText(_translate("MainWindow", "<html><head/><body><p><span style=\" font-size:10pt;\">Fit parameters</span></p></body></html>"))
        self.Label_npieces.setText(_translate("MainWindow", "<html><head/><body><p><span style=\" font-size:10pt;\">npieces</span></p></body></html>"))
        # self.Label_sig_lower.setText(_translate("MainWindow", "<html><head/><body><p><span style=\" font-size:10pt;\">sigma_lower</span></p></body></html>"))
        # self.Label_sig_upper.setText(_translate("MainWindow", "<html><head/><body><p><span style=\" font-size:10pt;\">sigma_upper</span></p></body></html>"))
        # self.Label_maxiter.setText(_translate("MainWindow", "<html><head/><body><p><span style=\" font-size:10pt;\">maxiter</span></p></body></html>"))
        self.Label_switch.setText(_translate("MainWindow", "<html><head/><body><p><span style=\" font-size:10pt;\">show_residual</span></p></body></html>"))
        self.Button_fit.setText(_translate("MainWindow", "Fit"))
    


# -*- coding: utf-8 -*-
import os
from copy import deepcopy

# NumPy
import numpy as np
# AstroPy
import astropy.units as u
from astropy.io import ascii, fits
from astropy.time import Time
from astropy.table import Table
# drpy
from drpy.onedspec.io import _Spectrum1D_to_hdu, loadSpectrum1D

from .__init__ import __version__ as version


def _plotSpectrum(ax, x, y, xlim, ylim, xlabel, ylabel):
    """Plot spectrum."""

    ax.step(x, y, 'k-', lw=0.8, where='mid')
    ax.tick_params(
        which='major', direction='in', top=True, right=True, length=5, width=1, 
        labelsize=12)
    ax.tick_params(
        which='minor', direction='in', top=True, right=True, length=3, width=1, 
        labelsize=12)
    ax.set_xlim(xlim)
    ax.set_ylim(ylim)
    ax.set_xlabel(xlabel, fontsize=12)
    ax.set_ylabel(ylabel, fontsize=12)

    return ax


def loadSpectrum(file_name, file_format, reverse):
    """Load uncalibrated spectrum.

    Parameters
    ----------
    file_name : str
        File name.

    file_format : str
        File format.

    reverse : bool
        Reverse ot not.

    Returns
    -------
    index : `~numpy.ndarray`
        Pixel index.

    count : `~numpy.ndarray`
        Normalized count.

    Raise 
    -----
    1) load file (empty)
    2) all nan or inf
    3) less than 5 points
    """

    # Load
    if file_format == 'fits':
        tbl = loadSpectrum1D(file_name, ext='spec')

        length = tbl.data.shape[-1]

        # Count
        count = tbl.data.flatten()[:length]

        # Unit of count
        unit_count = tbl.unit

    else:
        if file_format == 'ascii':
            tbl = ascii.read(file_name, guess=True, data_start=0)
        
        elif file_format == 'ecsv':
            tbl = Table.read(file_name, format='ascii.ecsv')
        
        length = len(tbl)

        # Count
        # Assume that flux is the second column for multi-column case.
        if len(tbl.colnames) > 1:
            colname = tbl.colnames[1]
        # Use the only column
        else:
            colname = tbl.colnames[0]
        count = tbl[colname].data

        # Unit of count
        if tbl[colname].unit is not None:
            unit_count = tbl[colname].unit
        # Default unit (use config in the future?)
        else:
            unit_count = u.Unit('')

    # Index
    index = np.arange(length)
    
    # Mask bad pixels
    mask = np.isfinite(count)
    if mask.any():
        count = np.interp(index, index[mask], count[mask])

    # Normalize
    count /= count.max()

    # Reverse
    if reverse:
        count = count[::-1]

    # Header
    if 'header' in tbl.meta:
        header = tbl.meta['header']

    else:
        header = fits.Header(tbl.meta)

    return index, count, unit_count, header


def saveSpectrum(file_name, spectrum, peak_table, saveECSV):
    """Save calibrated spectrum.

    Parameters
    ----------
    file_name : str
        File name.

    spectrum : `~astropy.table.table.Table`
        Calibrated spectrum to be saved.

    peak_table : `~astropy.table.table.Table`
        Properties of the peaks used to derive dispersion solution.

    saveECSV : bool
        If `True`, two .ecsv files containing the calibrated spectrum and the peak 
        properties are saved along with the .fits file.
    """
    
    hdu_spec = _Spectrum1D_to_hdu(spectrum, spectrum.meta['header'])
    
    hdu_spec.header['ORIGIN'] = (f'wcpy (version {version})', 'file generator')
    hdu_spec.header['DATE'] = (
        f'{Time.now().to_value("iso", subfmt="date_hm")}', 'date file was generated')

    hdu_peak = fits.table_to_hdu(peak_table)

    hdu_peak.header['ORIGIN'] = hdu_spec.header['ORIGIN']
    hdu_peak.header['DATE'] = hdu_spec.header['DATE']

    hdu_list = fits.HDUList([fits.PrimaryHDU(), hdu_spec, hdu_peak])

    hdu_list.writeto(file_name, overwrite=True)

    if saveECSV:
        Table.read(hdu_spec).write(
            os.path.splitext(file_name)[0] + '_spec.ecsv', format='ascii.ecsv', 
            overwrite=True)
        peak_table.write(
            os.path.splitext(file_name)[0] + '_peak.ecsv', format='ascii.ecsv', 
            overwrite=True)


# -*- coding: utf-8 -*-
# PyQt5
from PyQt5 import QtCore, QtGui, QtWidgets


class CheckBoxFileDialog(QtWidgets.QFileDialog):
    def __init__(self, check_box_text):
        super().__init__()
        self.setOption(QtWidgets.QFileDialog.DontUseNativeDialog)
        self.check_box = QtWidgets.QCheckBox(text=check_box_text)
        # Beautify
        self.layout().addWidget(self.check_box)


class RectSwitch(QtWidgets.QAbstractButton):
    def __init__(self, parent=None, margin=2, thumb_width=0.6, duration=150):
        super().__init__(parent=parent)

        self._margin = margin
        self._thumb_width = thumb_width
        self._duration = duration
        self.setColor()
        self.setCheckable(True)
        self.setFixedSize(self.sizeHint())


    def setColor(self):
        palette = self.palette()
        # Border color
        self._light_border_color = palette.light().color() # QColor
        self._dark_border_color = palette.shadow().color() # QColor
        # Face brush
        if self._margin < 0:
            self._track_brush = palette.shadow()  # QBrush
        else:
            self._track_brush = palette.light()   # QBrush
        self._thumb_brush = palette.button()      # QBrush
        # Opacity
        self._track_opacity = 1
        self._thumb_opacity = 1
        # Text
        self._text = {
            True: 'on',
            False: 'off',
        }
        self._text_color = {
            True: QtCore.Qt.black,             # QColor
            False: palette.mid().color()       # QColor
        }
        self._text_opacity = 1


    def sizeHint(self):
        return QtCore.QSize(80, 20)


    def setGeometry(self):
        self._track_width = min(self.width(), self.width() + 2 * self._margin)
        self._track_height = min(self.height(), self.height() + 2 * self._margin)
        self._thumb_width *= self.width()
        self._thumb_height = self._track_height - 2 * self._margin

        if self._margin > 0:
            self._base_offset = self._margin + self._thumb_width / 2
        else:
            self._base_offset = self._thumb_width / 2
        self._end_offset = {
            True: lambda: self.width() - self._base_offset,
            False: lambda: self._base_offset,
        }
        self._offset = self._base_offset


    def resizeEvent(self, event):
        super().resizeEvent(event)
        self.setGeometry()
        self.offset = self._end_offset[self.isChecked()]()


    @QtCore.Property(int)
    def offset(self):
        return self._offset


    @offset.setter
    def offset(self, value):
        self._offset = value
        self.update()


    # def setChecked(self, checked):
    #     super().setChecked(checked)
    #     self.offset = self._end_offset[checked]()


    def paintEvent(self, event):
        light_border_color = self._light_border_color
        dark_border_color = self._dark_border_color
        track_brush = self._track_brush
        thumb_brush = self._thumb_brush
        text_color = self._text_color[self.isEnabled()]
        track_opacity = self._track_opacity
        thumb_opacity = self._thumb_opacity
        text_opacity = self._text_opacity

        p = QtGui.QPainter(self)
        p.setRenderHint(QtGui.QPainter.Antialiasing, True)

        p.setBrush(track_brush)
        p.setOpacity(track_opacity)
        p.setPen(QtCore.Qt.NoPen)
        p.drawRect(
            self._base_offset - self._thumb_width / 2 - self._margin,
            self.height() / 2 - self._thumb_height / 2 - self._margin,
            self._track_width,
            self._track_height,
        )
        p.setBrush(thumb_brush)
        p.setOpacity(thumb_opacity)
        p.setPen(QtGui.QPen(light_border_color, 1.2, QtCore.Qt.SolidLine))
        p.drawRect(
            self.offset - self._thumb_width / 2,
            self.height() / 2 - self._thumb_height / 2,
            self._thumb_width,
            self._thumb_height,
        )
        p.setPen(QtGui.QPen(dark_border_color, 1.5, QtCore.Qt.SolidLine))
        p.drawLine(
            self.offset - self._thumb_width / 2 + 1.2,
            self.height() / 2 + self._thumb_height / 2 - 1,
            self.offset + self._thumb_width / 2 - 0.5,
            self.height() / 2 + self._thumb_height / 2 - 1,
        )
        p.drawLine(
            self.offset + self._thumb_width / 2,
            self.height() / 2 - self._thumb_height / 2 + 1.2,
            self.offset + self._thumb_width / 2,
            self.height() / 2 + self._thumb_height / 2 - 0.5,
        )
        p.setPen(text_color)
        p.setOpacity(text_opacity)
        font = p.font()
        font.setPixelSize(0.65 * self._thumb_height)
        p.setFont(font)
        p.drawText(
            QtCore.QRectF(
                self.offset - self._thumb_width / 2,
                self.height() / 2 - self._thumb_height / 2,
                self._thumb_width,
                self._thumb_height,
            ),
            QtCore.Qt.AlignCenter,
            self._text[self.isChecked()],
        )


    def mouseReleaseEvent(self, event):
        super().mouseReleaseEvent(event)
        if event.button() == QtCore.Qt.LeftButton:
            anim = QtCore.QPropertyAnimation(self, b'offset', self)
            anim.setDuration(self._duration)
            anim.setStartValue(self.offset)
            anim.setEndValue(self._end_offset[self.isChecked()]())
            anim.start()


    def enterEvent(self, event):
        self.setCursor(QtCore.Qt.PointingHandCursor)
        super().enterEvent(event)


if __name__ == '__main__':

    app = QtWidgets.QApplication([])

    # Thumb size < track size (Gitlab style)
    rs1 = RectSwitch(margin=2, thumb_width=0.6)
    rs1.toggled.connect(lambda c: print('toggled', c))
    rs1.clicked.connect(lambda c: print('clicked', c))
    rs1.pressed.connect(lambda: print('pressed'))
    rs1.released.connect(lambda: print('released'))
    rs2 = RectSwitch(margin=2, thumb_width=0.6)
    rs2.setEnabled(False)

    layout = QtWidgets.QHBoxLayout()
    layout.addWidget(rs1)
    layout.addWidget(rs2)
    widget = QtWidgets.QWidget()
    widget.setLayout(layout)
    widget.show()

    app.exec()


# -*- coding: utf-8 -*-
from .mainwindow import Ui_MainWindow
from .widgets import CheckBoxFileDialog, RectSwitch
from .fonts import table_font

__all__ = ['Ui_MainWindow', 'CheckBoxFileDialog', 'RectSwitch', 'table_font']